/*
#       lidsadm.c --- The Linux Intrusion Detection System Configuration Tool 
#       (C) Xie Huagang 1999 All rights reserved.
#       EMail complaints to xhg@gem.ncic.ac.cn.
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <linux/capability.h>


#define LIDS_NONE	0
#define LIDS_ADD	1
#define LIDS_DELETE	2
#define LIDS_LIST	3
#define LIDS_DELETE_ALL	4
#define LIDS_MK_PASSWD	5
#define LIDS_SWITCH	6
#define LIDS_INIT	7
#define LIDS_UPDATE	8
#define LIDS_VIEW	9

#define LIDS_READONLY	1
#define LIDS_APPEND	2
#define LIDS_IGNORE	3

#define LIDS_EXTEND	4

#define LIDS_CONF	"/etc/lids.conf"
#define LIDS_CONF_COMMENT  "#\n# Please do not modify this file by hand\n#\n"
#define LIDS_LOCKS      "/proc/sys/lids/locks"

typedef __u32 kernel_cap_t;

/* Me ? Paranoiac !? */

#define LIDS_MAGIC_1 0x154e6748
#define LIDS_MAGIC_2 0x68a52d68
#define LIDS_MAGIC_3 0xe68486c6
#define LIDS_MAGIC_4 0xd94aac1d

typedef int lids_flags_t;
typedef char passwd_t[64];

typedef struct lids_locks_s {
	int magic1;
	kernel_cap_t cap_bset;
	int magic2;
	lids_flags_t flags;
	int magic3;
	passwd_t passwd;
	int magic4;
} lids_locks_t;


#define CAP_FULL_SET (~0)
#define CAP_TO_MASK(flag) (1 << (flag))
#define cap_raise(cap, flag)  (cap |= CAP_TO_MASK(flag))
#define cap_lower(cap, flag)  (cap &= ~CAP_TO_MASK(flag))
#define cap_raised(cap, flag) (cap & CAP_TO_MASK(flag) & CAP_FULL_SET)
#define flag_raise(flags, flag)  (flags |= CAP_TO_MASK(flag))
#define flag_lower(flags, flag)  (flags &= ~CAP_TO_MASK(flag))

static char *thecaps[] =
{
	"CAP_CHOWN",            "chown(2)/chgrp(2)",
	"CAP_DAC_OVERRIDE",     "DAC access",
	"CAP_DAC_READ_SEARCH",  "DAC read",
	"CAP_FOWNER",           "owner ID not equal user ID",
	"CAP_FSETID",           "effective user ID not equal owner ID",
	"CAP_KILL",             "real/effective ID not equal process ID",
	"CAP_SETGID",           "setgid(2)",
	"CAP_SETUID",           "set*uid(2)",
	"CAP_SETPCAP",          "transfer capability",
	"CAP_LINUX_IMMUTABLE",  "immutable and append file attributes",
	"CAP_NET_BIND_SERVICE", "binding to ports below 1024",
	"CAP_NET_BROADCAST",    "broadcasting/listening to multicast",
	"CAP_NET_ADMIN",        "interface/firewall/routing changes",
	"CAP_NET_RAW",          "raw sockets",
	"CAP_IPC_LOCK",         "locking of shared memory segments",
	"CAP_IPC_OWNER",        "IPC ownership checks",
	"CAP_SYS_MODULE",       "insertion and removal of kernel modules",
	"CAP_SYS_RAWIO",        "ioperm(2)/iopl(2) access",
	"CAP_SYS_CHROOT",       "chroot(2)",
	"CAP_SYS_PTRACE",       "ptrace(2)",
	"CAP_SYS_PACCT",        "configuration of process accounting",
	"CAP_SYS_ADMIN",        "tons of admin stuff",
	"CAP_SYS_BOOT",         "reboot(2)",
	"CAP_SYS_NICE",         "nice(2)",
	"CAP_SYS_RESOURCE",     "setting resource limits",
	"CAP_SYS_TIME",         "setting system time",
	"CAP_SYS_TTY_CONFIG",   "tty configuration",
	NULL, NULL,
};
static char *theflags[] =
{
  "LIDS",	        	"LIDS itself",
  "INIT_CHILDREN_LOCK",         "protect any running init child from being killed",
  "RELOAD_CONF",                "reload config. file and inode/dev of special programs",
	NULL, NULL,
};

char *program;

struct lids {
	unsigned long int inode;
	unsigned int	  dev;
	int	 type;
	char 	 filename[1024];
};

struct lids lids[1024];
int	last;

void exit_error (int status, char *msg);
void exit_help();
void lids_add_file(char *filename,int type);
void lids_del_file(char *filename,int type);
void lids_del_all();
void lids_list_file(int type);
void lids_make_rmd160_passwd();
int   lids_verify();
void lids_switch(int optind, int argc, char *argv[]);
void lids_init(int optind, int argc, char *argv[]);
void lids_update();
#ifdef VIEW
void lids_view();
#endif
main(int argc,char **argv)
{
	int	command = LIDS_NONE;
	int 	type=LIDS_NONE;
	char 	filename[1024];
	char 	del_file[1024];
	char 	sw_flag[16];
	int	c;

	program=argv[0];
	if(getuid()!=0) {
		exit_error(2, "You must be root to run this program");
	}
#ifdef VIEW
	while ((c = getopt (argc, argv, "VUAD:ZLSIPr:a:z:eh")) != -1){
#else
	while ((c = getopt (argc, argv, "UAD:ZLSIPr:a:z:eh")) != -1){
#endif
        	switch (c)
			{
#ifdef VIEW
			case 'V':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_VIEW;
				break;
#endif
		        case 'U':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_UPDATE;
				break;
			case 'A':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_ADD;
				break;
			case 'D':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				strcpy(del_file,optarg);
				command = LIDS_DELETE;
				break;
			case 'L':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_LIST;
				break;
			case 'e':
				if (command != LIDS_LIST)
					exit_error (2, "error commands specified");
				type = LIDS_EXTEND;
				break;
			case 'Z':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_DELETE_ALL;
				break;
			case 'S':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_SWITCH;
				break;
			case 'I':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_INIT;
				break;
			case 'r':
			case 'a':
			case 'd':
			case 'z':
				if (command == LIDS_NONE)
					exit_error (2, "no comands specified");
				if ( command == LIDS_DELETE)
					exit_error (2, "error comands specified");
				if ( type != LIDS_NONE )
					exit_error (2, "multiple types specified");
				if( c == 'r' ) type = LIDS_READONLY;
				else if( c == 'a' ) type = LIDS_APPEND;
				else if( c == 'z' ) type = LIDS_IGNORE;
				strcpy(filename,optarg);
				break;
			case 'P':
				if (command != LIDS_NONE)
					exit_error (2, "multiple commands specified");
				command = LIDS_MK_PASSWD;
				break;
			case 'h':
			default:
				exit_help();	
			}
	}
	if ((command != LIDS_INIT) && (command != LIDS_SWITCH) && (optind < argc))
        	exit_error (2, "unknown arguments found on commandline");
	if ( command == LIDS_NONE || argc < 2 )
		exit_help();
	if ( command == LIDS_ADD & type == LIDS_NONE) 
		exit_help();
	switch(command) {
	case LIDS_ADD:
		printf("ADD\n");
		lids_add_file(filename,type);
		break;
	case LIDS_DELETE:
		printf("DELETE\n");
		lids_del_file(del_file,type);
		break;
	case LIDS_LIST:
		printf("LIST\n");
		lids_list_file(type);
		break;
	case LIDS_DELETE_ALL:
		printf("DELETE_ALL\n");
		lids_del_all();
		break;
	case LIDS_MK_PASSWD:
		printf("MAKE PASSWD\n");
		lids_make_rmd160_passwd();
		break;
	case LIDS_SWITCH:
		printf("SWITCH\n");
		lids_switch(optind,argc,argv);
		break;
	case LIDS_INIT:
		printf("INIT\n");
		lids_init(optind,argc,argv);
		break;
	case LIDS_UPDATE:
		printf("UPDATE\n");
		lids_update();
		break;
#ifdef VIEW
	case LIDS_VIEW:
		printf("VIEW\n");
		lids_view();
		break;
#endif
	}
	exit(0);
}

void lids_del_all(char *filename,int type)
{
	FILE *fd;
	
	if((fd = fopen(LIDS_CONF,"w"))== NULL ) {
		exit_error(3,"can not open conf file");
	}
	fwrite(LIDS_CONF_COMMENT,1,strlen(LIDS_CONF_COMMENT),fd);
	fclose(fd);
}
void lids_read_conf()
{
	FILE 	*fd;
	int	i=0,type;
	char 	buffer[1024];
	char    *p,*q;
	char 	ino_str[1024];
	char 	ino_dev[1024];
	char 	ino_filename[1024];
	
		
	if((fd = fopen(LIDS_CONF,"r"))== NULL ) {
		exit_error(3,"can not open conf file");
	}

	while(fgets(buffer,1024,fd) && i < 1024) {
		if ( buffer[0] == '#' ) continue;
		if ( buffer[strlen(buffer)-1] == '\n' )
			buffer[strlen(buffer)-1] = '\0' ;
		if ( buffer[0] == '[' ) {
			if ( strncmp(buffer,"[READONLY]",\
				     strlen("[READONLY]"))==0 ) 
				type = LIDS_READONLY;
			else if ( strncmp(buffer,"[APPEND]",8) ==0 )  
				type = LIDS_APPEND; 
			else 
				type = LIDS_IGNORE; 
			continue;
		}
		memset(ino_str,'\0',1024);
		memset(ino_dev,'\0',1024);
		memset(ino_filename,'\0',1024);
		p = q = NULL;
		p = strchr(buffer,':');	
		if ( p == NULL ) continue; 
		strncpy(ino_str,buffer,p-buffer);
		p++;
		q = strchr(p,':');
		if ( q == NULL ) continue; 
		strncpy(ino_dev,p,q-p);
		strcpy(ino_filename,q+1);

		lids[i].inode = atol(ino_str);
		lids[i].dev = atoi(ino_dev);
		lids[i].type = type;
		strcpy(lids[i].filename,ino_filename);
#ifdef DEBUG
		printf("inode = %d , dev = %d ,file=%s , type = %d \n",lids[i].inode,lids[i].dev,lids[i].filename,lids[i].type);
#endif
		i++;
	}
	last = i;

	fclose(fd);
}

void lids_write_file()
{
	int i;
	FILE *fd;
	
	if((fd = fopen(LIDS_CONF,"w"))== NULL ) {
		exit_error(3,"can not open conf file");
	}
	fwrite(LIDS_CONF_COMMENT,1,strlen(LIDS_CONF_COMMENT),fd);
	fprintf(fd,"[READONLY]\n");
	for(i=0;i<last;i++) {
		if ( lids[i].type == LIDS_READONLY ) 
			fprintf(fd,"%d:%d:%s\n",lids[i].inode,lids[i].dev,lids[i].filename);
	}
	fprintf(fd,"[APPEND]\n");
	for(i=0;i<last;i++) {
		if ( lids[i].type == LIDS_APPEND )
			fprintf(fd,"%d:%d:%s\n",lids[i].inode,lids[i].dev,lids[i].filename);
	}
	
	fprintf(fd,"[IGNORE]\n");
	for(i=0;i<last;i++) {
		if ( lids[i].type == LIDS_IGNORE )
			fprintf(fd,"%d:%d:%s\n",lids[i].inode,lids[i].dev,lids[i].filename);
	}

	fclose(fd);
}


void lids_list_file(int type)
{
	int i;
	lids_read_conf();
#ifdef DEBUG
	printf("last = %d \n",last);
#endif
	printf("\t \t  READ ONLY FILES \n");
	printf("\t \t  =============== \n");
	if ( type == LIDS_EXTEND) 
		printf("    Inode   \t       Dev     ");
	printf("\tFilename \n");
	printf("     --------------------------------------\n");
	for(i=0;i<last && lids[i].type == LIDS_READONLY;i++) {
		if( type == LIDS_EXTEND) {
			printf( "%10d \t%10d",
				lids[i].inode,lids[i].dev );
		}
		printf("\t%s\n",lids[i].filename);
	}	
	
	printf("\n\t\tAPPEND ONLY FILES \n");
	printf("\t\t================= \n");
	if ( type == LIDS_EXTEND) 
		printf("    Inode   \t       Dev     ");
	printf("\tFilename \n");
	printf("    --------------------------------------\n");
	for(i=0;i<last;i++) {
		if ( lids[i].type == LIDS_APPEND )  {
			if( type == LIDS_EXTEND) {
				printf( "%10d \t%10d",
					lids[i].inode,lids[i].dev );
			}
			printf("\t%s\n",lids[i].filename);
		}
		
	}
	printf("\n\t\tIgnore Protection \n");
	if ( type == LIDS_EXTEND) 
		printf("    Inode   \t       Dev     ");
	printf("\tFilename \n");
	printf("    --------------------------------------\n");
	for(i=0;i<last;i++) {
		if ( lids[i].type == LIDS_IGNORE)  {
			if( type == LIDS_EXTEND) {
				printf( "%10d \t%10d",
					lids[i].inode,lids[i].dev );
			}
			printf("\t%s\n",lids[i].filename);
		}
		
	}
	printf("\n\n");
}
/*
 *	lids_search_file()
 *
 *	lids search the file by given file and type.
 *
 */
int lids_search_file(char *filename) 
{
	int i ;

	for ( i = 0; i<last ; i++ ) {
		if ( strcmp(lids[i].filename , filename )== 0) {
			return i;
		}
	}

	return -1;
}

/*
 *	lids_search_inode()
 *
 *	lids search the file by given inode , dev  and type.
 *
 */
int lids_search_inode(struct stat st) 
{
	int i;

	for ( i = 0; i<last ; i++ ) {
		if ( lids[i].inode == st.st_ino && 
		     lids[i].dev == st.st_dev ) {   
			return i;
		}
	}	
	return -1;
}
void lids_del_file(char *filename,int type)
{
	int i;
	lids_read_conf();

	i = lids_search_file(filename);
		
	if(i< 0  )
		exit_error(2,"the file does not exit in the conf file");
	lids[i].type = -1;

	lids_write_file();
}
void lids_add_file(char *filename,int type)
{
	struct stat st;

	lids_read_conf();

        if ( last >= 1024 ) {
                exit_error(2, "cannot exceed 1024 entries.");
        }

	if ( !stat(filename,&st) ) {
		if ( lids_search_inode(st) !=-1 )
			exit_error(2," the file which has the same inode and dev has been exited");
		lids[last].inode= st.st_ino;
		lids[last].dev = st.st_dev;
		lids[last].type = type;

		strcpy(lids[last].filename,filename);
		
		last++;
		lids_write_file();
	}
	else {
		exit_error(2, " can not find the file");
	}
}

void lids_update()
{
	char type[30]="read-only\0append\0\0\0\0ignore\0\0\0\0";
	struct stat st;
	int i;
	lids_read_conf();
	

	for (i=0; i<last; i++) {
		if (stat(lids[i].filename,&st)) {
			printf("%s doesn't exists anymore. Not removed.\n",lids[i].filename);
		}
		else {
			if ((st.st_ino != lids[i].inode) || (st.st_dev != lids[i].dev)) {
				printf("%s (type:%s) was (%u:%u inode %lu) instead of (%u:%u %lu). Corrected.\n",
				       lids[i].filename,type+10*(lids[i].type-1),
				       MAJOR(lids[i].dev),MINOR(lids[i].dev),lids[i].inode,
				       MAJOR((unsigned long)st.st_dev),MINOR((unsigned long)st.st_dev),st.st_ino);
				lids[i].inode = st.st_ino;
				lids[i].dev = st.st_dev;
			}
		}
	}
	lids_write_file();
}


void
exit_error (int status, char *msg)
{
	fprintf (stderr, "%s: %s\n", program, msg);
	if(status == 3 )
		perror("reason:");
	printf("\n");
	exit (status);
}
void
exit_help ()
{
	int i;
	printf ("%s v0.4.4 for Linux Intrusion Detection System\n" 
		"	       	Xie Huagang<xhg@gem.ncic.ac.cn>\n"
		"	        Philippe Biondi <philippe.biondi@webmotion.net>\n\n"
		"Usage: %s -A -[a|r|d|z] [file|diretory|device]\n"
		"       %s -D  file/directory\n"
		"       %s -Z\n"
		"       %s -U\n"
		"       %s -L\n"
		"       %s -P\n"
		"       %s -[S|I] -- [+|-][CAPABILITY|LIDS_FLAG] [...]\n"
#ifdef VIEW
		"       %s -V\n"
#endif
		"       %s -h\n\n"
		"Commands:\n"
		"       -A  To add an entry\n"
		"       -D  To delete an entry\n"
		"       -Z  To delete all entries \n"
		"       -U  To update dev/inode numbers\n"
		"       -L  To list all entries \n"
		"       -P  To encrypt a password with RipeMD-160\n"
		"       -S  To submit a password to switch some protections\n"
		"       -I  To switch some protections without submitting password (sealing time)\n"
#ifdef VIEW
		"       -V  To view current LIDS state (caps/flags)\n"
#endif
		"       -h  To list this help \n"
	
		"Types:\n"
		"	    -r	Read Only \n"
		"	    -a  APPEND Only \n"
		"	    -z  Ignore protection \n"
#ifdef VIEW
		,program
#endif
		,program,program,program,program,program,program,program,program,program);

	printf("\nAvailable capabilities:\n");
	for (i=0; thecaps[2*i] != NULL; i++) 
		printf("%20s %s\n",thecaps[2*i],thecaps[2*i+1]);
	printf("\nAvailable flags:\n");
	for (i=0; theflags[2*i] != NULL; i++) 
		printf("%20s %s\n",theflags[2*i],theflags[2*i+1]);

	exit(1);
}
void lids_make_rmd160_passwd()
{
	char passwd[64];
	int times=0;
	int ok;
	
	while(times < 3) {
		times++;	
		if ((ok=read_rmd160_passwd(passwd,1,2))==0)  
			break;
	}
	if (ok==0)
		printf("RipeMD-160 encrypted password : %s\n",passwd); 
}

	

void lids_set_caps(int optind, int argc, char *argv[],lids_locks_t *locks)

{
	int i;
	
	for (i=optind; i < argc; i++) {
		int capnum=-1;
		char *capstr=NULL, *capdescstr=NULL;
		int flagnum=-1;
		char *flagstr=NULL, *flagdescstr=NULL;
		int j=0;
			
		while (thecaps[j] != NULL) {
			if (!strcmp(argv[i]+1, thecaps[j])) {
				capnum = j/2;
				capstr = thecaps[j];
				capdescstr = thecaps[j+1];
				break;
			}
			j += 2;
		}
		
		if (capnum == -1) {
			j=0;
			
			while (theflags[j] != NULL) {
				if (!strcmp(argv[i]+1, theflags[j])) {
					flagnum = j/2;
					flagstr = theflags[j];
					flagdescstr = theflags[j+1];
					break;
				}
				j += 2;
			}
		}
			
		if ( ((capnum == -1) || (capstr == NULL)) &&
		     ((flagnum == -1) || (flagstr == NULL)) ) {
			fprintf(stderr, "    %s: invalid capability/flag\n", argv[i]);
			exit(1);
		}

		switch (argv[i][0]) {
		case '+':
			if (flagnum == -1)
				cap_raise(locks->cap_bset, capnum);
			else
				flag_raise(locks->flags, flagnum);
			break;
		case '-':
			if (flagnum == -1)
				cap_lower(locks->cap_bset, capnum);
			else
				flag_lower(locks->flags, flagnum);
			break;
		default:
			fprintf(stderr, "%s: invalid option\n", argv[i]);
			exit(1);
		}
			
	}
}

void lids_switch(int optind, int argc, char *argv[])
{
	int lk ;
	char passwd[BUFSIZ];
	lids_locks_t locks;

	if ((lk=open(LIDS_LOCKS,O_RDWR)) == -1) {
		perror("open");
		exit_error (2, "can't open " LIDS_LOCKS);
	}
	if (read(lk,&locks,sizeof(lids_locks_t))==-1) {
		perror("read");
		exit_error (2, "can't read " LIDS_LOCKS);
	}

	lids_set_caps(optind,argc,argv,&locks);

	locks.magic1=LIDS_MAGIC_1;
	locks.magic2=LIDS_MAGIC_2;
	locks.magic3=LIDS_MAGIC_3;
	locks.magic4=LIDS_MAGIC_4;

	read_rmd160_passwd(passwd,0,1);  

	strncpy(locks.passwd,passwd,64);

	if (write(lk,&locks,sizeof(lids_locks_t))==-1) {
		perror("write");
		exit_error (2, "can't write " LIDS_LOCKS);
	}
	close(lk);

}

void lids_init(int optind, int argc, char *argv[])
{
	int fd ;
	lids_locks_t locks;
	locks.cap_bset=0;
	locks.flags=0;
	
	if ((fd=open(LIDS_LOCKS,O_RDWR)) == -1) {
		perror("open");
		exit_error (2, "can't open " LIDS_LOCKS);
	}
	if (read(fd,&locks,sizeof(lids_locks_t))==-1) {
		perror("read");
		exit_error (2, "can't read " LIDS_LOCKS);
	}

	lids_set_caps(optind,argc,argv,&locks);
	
	locks.magic1=LIDS_MAGIC_1;
	locks.magic2=LIDS_MAGIC_2;
	locks.magic3=LIDS_MAGIC_3;
	locks.magic4=LIDS_MAGIC_4;
	locks.passwd[0]=0;

	if (write(fd,&locks,sizeof(lids_locks_t))==-1) {
		perror("write");
		exit_error (2, "can't write " LIDS_LOCKS);
	}
	close(fd);
}
	
#ifdef VIEW
void lids_view()
{
	int fd ;
	lids_locks_t locks;
	int i;
	locks.cap_bset=0;
	locks.flags=0;

	if ((fd=open(LIDS_LOCKS,O_RDWR)) == -1) {
		perror("open");
		exit_error (2, "can't open " LIDS_LOCKS);
	}
	if (read(fd,&locks,sizeof(lids_locks_t))==-1) {
		perror("read");
		exit_error (2, "can't read " LIDS_LOCKS);
	}
	close(fd);

	for(i=0;(i < 8*sizeof(kernel_cap_t)) && (thecaps[2*i] != NULL);i++)
		    printf("%30s %i\n",thecaps[2*i],(locks.cap_bset >> i) & 1);
	for(i=0;(i < 8*sizeof(lids_flags_t)) && (theflags[2*i] != NULL);i++) 
		    printf("%30s %i\n",theflags[2*i],(locks.flags >> i) & 1);

}
#endif







